@@ -7,11 +7,11 @@
 //! # extern crate programinduction;
 //! use programinduction::lambda::{task_by_evaluation, Language, SimpleEvaluator};
 //!
-//! fn evaluate(name: &str, inps: &[i32]) -> i32 {
+//! fn evaluate(name: &str, inps: &[i32]) -> Result<i32, ()> {
 //!     match name {
-//!         "0" => 0,
-//!         "1" => 1,
-//!         "+" => inps[0] + inps[1],
+//!         "0" => Ok(0),
+//!         "1" => Ok(1),
+//!         "+" => Ok(inps[0] + inps[1]),
 //!         _ => unreachable!(),
 //!     }
 //! }
@@ -62,6 +62,14 @@ pub struct Language {
     pub primitives: Vec<(String, TypeSchema, f64)>,
     pub invented: Vec<(Expression, TypeSchema, f64)>,
     pub variable_logprob: f64,
+    /// Symmetry breaking prevents certain productions from being made. Specifically, an item `(f,
+    /// i, a)` means that enumeration will not yield an application of `f` where the `i`th argument
+    /// is `a`. This vec must be kept sorted — use via [`add_symmetry_violation`] and
+    /// [`violates_symmetry`].
+    ///
+    /// [`add_symmetry_violation`]: #method.add_symmetry_violation
+    /// [`violates_symmetry`]: #method.violates_symmetry
+    pub symmetry_violations: Vec<(usize, usize, usize)>,
 }
 impl Language {
     /// A uniform distribution over primitives and invented expressions, as well as the abstraction
@@ -75,6 +83,7 @@ impl Language {
             primitives,
             invented: vec![],
             variable_logprob: 0f64,
+            symmetry_violations: Vec::new(),
         }
     }
 
@@ -121,6 +130,9 @@ impl Language {
     ///
     /// # Examples
     ///
+    /// The following example can be made more effective using the approach shown with
+    /// [`add_symmetry_violation`].
+    ///
     /// ```
     /// # #[macro_use] extern crate polytype;
     /// # extern crate programinduction;
@@ -131,7 +143,6 @@ impl Language {
     ///     ("0", ptp!(int)),
     ///     ("1", ptp!(int)),
     ///     ("+", ptp!(@arrow[tp!(int), tp!(int), tp!(int)])),
-    ///     (">", ptp!(@arrow[tp!(int), tp!(int), tp!(bool)])),
     /// ]);
     /// let exprs: Vec<Expression> = dsl.enumerate(ptp!(int))
     ///     .take(8)
@@ -153,6 +164,8 @@ impl Language {
     /// );
     /// # }
     /// ```
+    ///
+    /// [`add_symmetry_violation`]: #method.add_symmetry_violation
     pub fn enumerate(&self, tp: TypeSchema) -> Box<Iterator<Item = (Expression, f64)>> {
         let (tx, rx) = bounded(1);
         let dsl = self.clone();
@@ -184,7 +197,7 @@ impl Language {
     /// let dsl = circuits::dsl();
     /// let tasks = circuits::make_tasks(100);
     /// let ec_params = ECParams {
-    ///     frontier_limit: 5,
+    ///     frontier_limit: 10,
     ///     search_limit_timeout: Some(std::time::Duration::new(2, 0)),
     ///     search_limit_description_length: None,
     /// };
@@ -221,11 +234,11 @@ impl Language {
     /// # extern crate programinduction;
     /// use programinduction::lambda::{Language, SimpleEvaluator};
     ///
-    /// fn evaluate(name: &str, inps: &[i32]) -> i32 {
+    /// fn evaluate(name: &str, inps: &[i32]) -> Result<i32, ()> {
     ///     match name {
-    ///         "0" => 0,
-    ///         "1" => 1,
-    ///         "+" => inps[0] + inps[1],
+    ///         "0" => Ok(0),
+    ///         "1" => Ok(1),
+    ///         "+" => Ok(inps[0] + inps[1]),
     ///         _ => unreachable!(),
     ///     }
     /// }
@@ -243,10 +256,10 @@ impl Language {
     /// assert_eq!(evaluated, 8);
     /// # }
     /// ```
-    pub fn eval<E, V>(&self, expr: &Expression, evaluator: E, inps: &[V]) -> Option<V>
+    pub fn eval<V, E>(&self, expr: &Expression, evaluator: E, inps: &[V]) -> Result<V, E::Error>
     where
-        E: Evaluator<Space = V>,
         V: Clone + PartialEq + Send + Sync,
+        E: Evaluator<Space = V>,
     {
         eval::eval(self, expr, &Arc::new(evaluator), inps)
     }
@@ -254,10 +267,15 @@ impl Language {
     /// Like [`eval`], but useful in settings with a shared evaluator.
     ///
     /// [`eval`]: #method.eval
-    pub fn eval_arc<E, V>(&self, expr: &Expression, evaluator: &Arc<E>, inps: &[V]) -> Option<V>
+    pub fn eval_arc<V, E>(
+        &self,
+        expr: &Expression,
+        evaluator: &Arc<E>,
+        inps: &[V],
+    ) -> Result<V, E::Error>
     where
-        E: Evaluator<Space = V>,
         V: Clone + PartialEq + Send + Sync,
+        E: Evaluator<Space = V>,
     {
         eval::eval(self, expr, evaluator, inps)
     }
@@ -323,6 +341,98 @@ impl Language {
         Ok(self.invented.len() - 1)
     }
 
+    /// Introduce a symmetry-breaking pattern to the Language.
+    ///
+    /// # Examples
+    ///
+    /// ```
+    /// # #[macro_use] extern crate polytype;
+    /// # extern crate programinduction;
+    /// # use programinduction::lambda::{Expression, Language};
+    /// # fn main() {
+    /// let mut dsl = Language::uniform(vec![
+    ///     ("0", ptp!(int)),
+    ///     ("1", ptp!(int)),
+    ///     ("+", ptp!(@arrow[tp!(int), tp!(int), tp!(int)])),
+    /// ]);
+    /// // disallow (+ 0 _) and (+ _ 0)
+    /// dsl.add_symmetry_violation(2, 0, 0);
+    /// dsl.add_symmetry_violation(2, 1, 0);
+    /// // disallow (+ (+ ..) _), so effort isn't wasted with (+ _ (+ ..))
+    /// dsl.add_symmetry_violation(2, 0, 2);
+    ///
+    /// let exprs: Vec<Expression> = dsl.enumerate(ptp!(int))
+    ///     .take(6)
+    ///     .map(|(expr, _log_prior)| expr)
+    ///     .collect();
+    ///
+    /// // enumeration can be far more effective with symmetry-breaking:
+    /// assert_eq!(
+    ///     exprs,
+    ///     vec![
+    ///         Expression::Primitive(0),
+    ///         Expression::Primitive(1),
+    ///         dsl.parse("(+ 1 1)").unwrap(),
+    ///         dsl.parse("(+ 1 (+ 1 1))").unwrap(),
+    ///         dsl.parse("(+ 1 (+ 1 (+ 1 1)))").unwrap(),
+    ///         dsl.parse("(+ 1 (+ 1 (+ 1 (+ 1 1))))").unwrap(),
+    ///     ]
+    /// );
+    /// # }
+    /// ```
+    pub fn add_symmetry_violation(&mut self, primitive: usize, arg_index: usize, arg: usize) {
+        let x = (primitive, arg_index, arg);
+        if let Err(i) = self.symmetry_violations.binary_search(&x) {
+            self.symmetry_violations.insert(i, x)
+        }
+    }
+    /// Check whether expressions break symmetry.
+    ///
+    /// # Examples
+    ///
+    /// ```
+    /// # #[macro_use] extern crate polytype;
+    /// # extern crate programinduction;
+    /// # use programinduction::lambda::{Expression, Language};
+    /// # fn main() {
+    /// let mut dsl = Language::uniform(vec![
+    ///     ("0", ptp!(int)),
+    ///     ("1", ptp!(int)),
+    ///     ("+", ptp!(@arrow[tp!(int), tp!(int), tp!(int)])),
+    /// ]);
+    /// dsl.add_symmetry_violation(2, 0, 0);
+    /// dsl.add_symmetry_violation(2, 1, 0);
+    /// dsl.add_symmetry_violation(2, 0, 2);
+    ///
+    /// let f = &Expression::Primitive(2); // +
+    /// let x = &Expression::Primitive(0); // 0
+    /// assert!(dsl.violates_symmetry(f, 0, x));
+    /// let x = &dsl.parse("(+ 1 1)").unwrap();
+    /// assert!(dsl.violates_symmetry(f, 0, x));
+    /// assert!(!dsl.violates_symmetry(f, 1, x));
+    /// # }
+    pub fn violates_symmetry(&self, f: &Expression, index: usize, x: &Expression) -> bool {
+        match (f, x) {
+            (&Expression::Primitive(f), &Expression::Primitive(x)) => {
+                let x = (f, index, x);
+                self.symmetry_violations.binary_search(&x).is_ok()
+            }
+            (&Expression::Primitive(f), &Expression::Application(ref x, _)) => {
+                let mut z: &Expression = &**x;
+                while let Expression::Application(ref x, _) = *z {
+                    z = x
+                }
+                if let Expression::Primitive(x) = *z {
+                    let x = (f, index, x);
+                    self.symmetry_violations.binary_search(&x).is_ok()
+                } else {
+                    false
+                }
+            }
+            _ => false,
+        }
+    }
+
     /// Remove all invented expressions by pulling out their underlying expressions.
     pub fn strip_invented(&self, expr: &Expression) -> Expression {
         expr.strip_invented(&self.invented)
@@ -468,10 +578,7 @@ impl Expression {
             Expression::Primitive(num) => if let Some(prim) = dsl.primitives.get(num as usize) {
                 Ok(prim.1.clone().instantiate_owned(ctx))
             } else {
-                Err(InferenceError::BadExpression(format!(
-                    "primitive does not exist: {}",
-                    num
-                )))
+                Err(InferenceError::InvalidPrimitive(num))
             },
             Expression::Application(ref f, ref x) => {
                 let f_tp = f.infer(dsl, &mut ctx, env, indices)?;
@@ -506,10 +613,7 @@ impl Expression {
             Expression::Invented(num) => if let Some(inv) = dsl.invented.get(num as usize) {
                 Ok(inv.1.clone().instantiate_owned(ctx))
             } else {
-                Err(InferenceError::BadExpression(format!(
-                    "invention does not exist: {}",
-                    num
-                )))
+                Err(InferenceError::InvalidInvention(num))
             },
         }
     }
@@ -629,11 +733,11 @@ impl Expression {
 /// # extern crate programinduction;
 /// use programinduction::lambda::{task_by_evaluation, Language, SimpleEvaluator};
 ///
-/// fn evaluate(name: &str, inps: &[i32]) -> i32 {
+/// fn evaluate(name: &str, inps: &[i32]) -> Result<i32, ()> {
 ///     match name {
-///         "0" => 0,
-///         "1" => 1,
-///         "+" => inps[0] + inps[1],
+///         "0" => Ok(0),
+///         "1" => Ok(1),
+///         "+" => Ok(inps[0] + inps[1]),
 ///         _ => unreachable!(),
 ///     }
 /// }
@@ -664,8 +768,8 @@ where
     let evaluator = Arc::new(evaluator);
     let oracle = Box::new(move |dsl: &Language, expr: &Expression| {
         let success = examples.iter().all(|&(ref inps, ref out)| {
-            if let Some(ref o) = dsl.eval_arc(expr, &evaluator, inps) {
-                o == out
+            if let Ok(o) = dsl.eval_arc(expr, &evaluator, inps) {
+                o == *out
             } else {
                 false
             }
@@ -746,7 +850,8 @@ impl<T: Clone> Index<usize> for LinkedList<T> {
 
 #[derive(Debug, Clone)]
 pub enum InferenceError {
-    BadExpression(String),
+    InvalidPrimitive(usize),
+    InvalidInvention(usize),
     Unify(UnificationError),
 }
 impl From<UnificationError> for InferenceError {
@@ -757,7 +862,8 @@ impl From<UnificationError> for InferenceError {
 impl fmt::Display for InferenceError {
     fn fmt(&self, f: &mut fmt::Formatter) -> Result<(), fmt::Error> {
         match *self {
-            InferenceError::BadExpression(ref msg) => write!(f, "invalid expression: '{}'", msg),
+            InferenceError::InvalidPrimitive(n) => write!(f, "primitive {} not in Language", n),
+            InferenceError::InvalidInvention(n) => write!(f, "invention {} not in Language", n),
             InferenceError::Unify(ref err) => write!(f, "could not unify to infer type: {}", err),
         }
     }
